/* replace normal networking commands with those in libtermnet, a bit
 * like termnet.h but using real functions rather than CPP macros
 */

/* note that we don't particularly care about types here, so everything
 * comes down to either a basic type or a void*
 */
#include <stdio.h>
int term_accept(int s,void *addr,int *addrlen);
int x__accept(int s,void *addr,int *addrlen);
int accept(int s,void *addr,int *addrlen)
{
  static int recursive=0;
  int retval=recursive++?x__accept(s,addr,addrlen):term_accept(s,addr,addrlen);
  /* fprintf(stderr,"XXX(%d):accept\n",recursive); */
  recursive--;
  return retval;
}

int term_bind(int s,void *name,int namelen);
int x__bind(int s,void *name,int namelen);
int bind(int s,void *name,int namelen)
{
  static int recursive=0;
  int retval=recursive++?x__bind(s,name,namelen):term_bind(s,name,namelen);
  /* fprintf(stderr,"XXX(%d):bind\n",recursive); */
  recursive--;
  return retval;
}

int term_chroot(const char *path);
int x__chroot(const char *path);
int chroot(const char *path)
{ 
  static int recursive=0;
  int retval=recursive++?x__chroot(path):term_chroot(path);
  /* fprintf(stderr,"XXX(%d):chroot\n",recursive); */
  recursive--;
  return retval;
}

int term_close(int fd);
int x__close(int fd);
int close(int fd)
{
  static int recursive=0;
  int retval=recursive++?x__close(fd):term_close(fd);
  /* fprintf(stderr,"XXX(%d):close\n",recursive); */
  recursive--;
  return retval;
}

int term_connect(int s,void *name,int namelen);
int x__connect(int s,void *name,int namelen);
int connect(int s,void *name,int namelen)
{
  static int recursive=0;
  int retval=recursive++?x__connect(s,name,namelen):term_connect(s,name,namelen);
  /* fprintf(stderr,"XXX(%d):connect\n",recursive); */
  recursive--;
  return retval;
}

int term_dup(int fd);
int x__dup(int fd);
int dup(int fd)
{
  static int recursive=0;
  int retval=recursive++?x__dup(fd):term_dup(fd);
  /* fprintf(stderr,"XXX(%d):dup\n",recursive); */
  recursive--;
  return retval;
}

int term_dup2(int fd1,int fd2);
int x__dup2(int fd1,int fd2);
int dup2(int fd1,int fd2)
{
  static int recursive=0;
  int retval=recursive++?x__dup2(fd1,fd2):term_dup2(fd1,fd2);
  /* fprintf(stderr,"XXX(%d):dup2\n",recursive); */
  recursive--;
  return retval;
}

int term_fcntl(int fd,int cmd,void *arg);
int x__fcntl(int fd,int cmd,void *arg);
int fcntl(int fd,int cmd,void *arg)
{
  static int recursive=0;
  int retval=recursive++?x__fcntl(fd,cmd,arg):term_fcntl(fd,cmd,arg);
  /* fprintf(stderr,"XXX(%d):fcntl\n",recursive); */
  recursive--;
  return retval;
}

int term_fork();
int x__fork();
int fork()
{
  static int recursive=0;
  int retval=recursive++?x__fork():term_fork();
  /* fprintf(stderr,"XXX(%d):fork\n",recursive); */
  recursive--;
  return retval;
}

void *term_gethostbyname(const char *name);
void *x__gethostbyname(const char *name);
void *gethostbyname(const char *name)
{
  static int recursive=0;
  void *retval=recursive++?x__gethostbyname(name):term_gethostbyname(name);
  /* fprintf(stderr,"XXX(%d):gethostbyname\n",recursive); */
  recursive--;
  return retval;
}

void *term_gethostbyaddr(const char *addr,int len,int type);
void *x__gethostbyaddr(const char *addr,int len,int type);
void *gethostbyaddr(const char *addr,int len,int type)
{
  static int recursive=0;
  void *retval=recursive++?x__gethostbyaddr(addr,len,type):term_gethostbyaddr(addr,len,type);
  /* fprintf(stderr,"XXX(%d):gethostbyaddr\n",recursive); */
  recursive--;
  return retval;
}

int term_gethostname(char *name,int namelen);
int x__gethostname(char *name,int namelen);
int gethostname(char *name,int namelen)
{
  static int recursive=0;
  int retval=recursive++?x__gethostname(name,namelen):term_gethostname(name,namelen);
  /* fprintf(stderr,"XXX(%d):gethostbyname\n",recursive); */
  recursive--;
  return retval;
}

int term_getpeername(int s,void *name,int *namelen);
int x__getpeername(int s,void *name,int *namelen);
int getpeername(int s,void *name,int *namelen)
{
  static int recursive=0;
  int retval=recursive++?x__getpeername(s,name,namelen):term_getpeername(s,name,namelen);
  /* fprintf(stderr,"XXX(%d):getpeername\n",recursive); */
  recursive--;
  return retval;
}

int term_getsockname(int s,void *name,int *namelen);
int x__getsockname(int s,void *name,int *namelen);
int getsockname(int s,void *name,int *namelen)
{
  static int recursive=0;
  int retval=recursive++?x__getsockname(s,name,namelen):term_getsockname(s,name,namelen);
  /* fprintf(stderr,"XXX(%d):getsockname\n",recursive); */
  recursive--;
  return retval;
}

int term_listen(int s,int backlog);
int x__listen(int s,int backlog);
int listen(int s,int backlog)
{
  static int recursive=0;
  int retval=recursive++?x__listen(s,backlog):term_listen(s,backlog);
  /* fprintf(stderr,"XXX(%d):listen\n",recursive); */
  recursive--;
  return retval;
}

void term_perror(const char *s);
void x__perror(const char *s);
void perror(const char *s)
{
  static int recursive=0;
  if(recursive++)
    x__perror(s);
  else
    perror(s);
  /* fprintf(stderr,"XXX(%d):perror\n",recursive); */
  recursive--;
}

int term_rcmd(char **ahost,unsigned short inport,char *luser,
         char *ruser,char *cmd,int *fd2p);
int x__rcmd(char **ahost,unsigned short inport,char *luser,
         char *ruser,char *cmd,int *fd2p);
int rcmd(char **ahost,unsigned short inport,char *luser,
         char *ruser,char *cmd,int *fd2p)
{
  static int recursive=0;
  int retval=recursive++?x__rcmd(ahost,inport,luser,ruser,cmd,fd2p):term_rcmd(ahost,inport,luser,ruser,cmd,fd2p);
  /* fprintf(stderr,"XXX(%d):rcmd\n",recursive); */
  recursive--;
  return retval;
}

int term_recv(int s,char *buf,int len,int flags);
int x__recv(int s,char *buf,int len,int flags);
int recv(int s,char *buf,int len,int flags)
{
  static int recursive=0;
  int retval=recursive++?x__recv(s,buf,len,flags):term_recv(s,buf,len,flags);
  /* fprintf(stderr,"XXX(%d):recv\n",recursive); */
  recursive--;
  return retval;
}

int term_recvfrom(int s,char *buf,int len,int flags,void *from,int *fromlen);
int x__recvfrom(int s,char *buf,int len,int flags,void *from,int *fromlen);
int recvfrom(int s,char *buf,int len,int flags,void *from,int *fromlen)
{
  static int recursive=0;
  int retval=recursive++?x__recvfrom(s,buf,len,flags,from,fromlen):term_recvfrom(s,buf,len,flags,from,fromlen);
  /* fprintf(stderr,"XXX(%d):recvfrom\n",recursive); */
  recursive--;
  return retval;
}

int term_send(int s,char *buf,int len,int flags);
int x__send(int s,char *buf,int len,int flags);
int send(int s,char *buf,int len,int flags)
{
  static int recursive=0;
  int retval=recursive++?x__send(s,buf,len,flags):term_send(s,buf,len,flags);
  /* fprintf(stderr,"XXX(%d):send\n",recursive); */
  recursive--;
  return retval;
}

int term_sendto(int s,char *buf,int len,int flags,void *to,int tolen);
int x__sendto(int s,char *buf,int len,int flags,void *to,int tolen);
int sendto(int s,char *buf,int len,int flags,void *to,int tolen)
{
  static int recursive=0;
  int retval=recursive++?x__sendto(s,buf,len,flags,to,tolen):term_sendto(s,buf,len,flags,to,tolen);
  /* fprintf(stderr,"XXX(%d):sendto\n",recursive); */
  recursive--;
  return retval;
}

int term_shutdown(int s,int how);
int x__shutdown(int s,int how);
int shutdown(int s,int how)
{
  static int recursive=0;
  int retval=recursive++?x__shutdown(s,how):term_shutdown(s,how);
  /* fprintf(stderr,"XXX(%d):shutdown\n",recursive); */
  recursive--;
  return retval;
}

int term_socket(int domain,int type,int protocol);
int x__socket(int domain,int type,int protocol);
int socket(int domain,int type,int protocol)
{
  static int recursive=0;
  int retval=recursive++?x__socket(domain,type,protocol):term_socket(domain,type,protocol);
  /* fprintf(stderr,"XXX(%d):socket\n",recursive); */
  recursive--;
  return retval;
}

#ifndef NO_STRERROR
char *term_strerror(int errnum);
char *x__strerror(int errnum);
char *strerror(int errnum)
{
  static int recursive=0;
  char *retval=recursive++?x__strerror(errnum):term_strerror(errnum);
  /* fprintf(stderr,"XXX(%d):strerror\n",recursive); */
  recursive--;
  return retval;
}
#endif

int term_vfork();
int x__vfork();
int vfork()
{
  static int recursive=0;
  int retval=recursive++?x__vfork():term_vfork();
  /* fprintf(stderr,"XXX(%d):vfork\n",recursive); */
  recursive--;
  return retval;
}
